from typing import Any, Collection, List

from fractal.core.specifications.generic.specification import Specification


class NotSpecification(Specification):
    def __init__(self, specification: Specification):
        self.specification = specification

    def is_satisfied_by(self, obj: Any) -> bool:
        return not self.specification.is_satisfied_by(obj)

    def to_collection(self) -> Collection:
        return [self.specification]

    def __str__(self):
        return f"{self.__class__.__name__}({self.specification})"

    def __eq__(self, other):
        return (
            isinstance(other, NotSpecification)
            and self.specification == other.specification
        )


class FieldValueSpecification(Specification):
    def __init__(self, field: str, value: Any):
        self.field = field
        self.value = value

    def __str__(self):
        return f"{self.__class__.__name__}({self.field}={self.value})"

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self.field == other.field
            and self.value == other.value
        )

    def to_collection(self) -> Collection:
        return {self.field, self.value}


class InSpecification(FieldValueSpecification):
    def __init__(self, field: str, values: List[Any]):
        super(InSpecification, self).__init__(field, values)

    def is_satisfied_by(self, obj: Any) -> bool:
        return getattr(obj, self.field) in self.value


class EqualsSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return getattr(obj, self.field) == self.value


class LessThanSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return getattr(obj, self.field) < self.value


class LessThanEqualSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return getattr(obj, self.field) <= self.value


class GreaterThanSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return getattr(obj, self.field) > self.value


class GreaterThanEqualSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return getattr(obj, self.field) >= self.value


class ContainsSpecification(FieldValueSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return self.value in getattr(obj, self.field)


class RegexStringMatchSpecification(ContainsSpecification):
    pass



import pytest


@pytest.fixture
def sqlalchemy_specification_builder():
    from fractal.contrib.sqlalchemy.specifications import (
        SqlAlchemyOrmSpecificationBuilder,
    )

    return SqlAlchemyOrmSpecificationBuilder



def test_is_healthy(inmemory_repository):
    assert inmemory_repository.is_healthy()


def test_add(inmemory_repository, an_object):
    inmemory_repository.add(an_object)

    assert len(inmemory_repository.entities) == 1
    assert list(inmemory_repository.entities.values())[0] == an_object


def test_update(inmemory_repository, an_object):
    inmemory_repository.add(an_object)

    assert list(inmemory_repository.entities.values())[0].name != "update"

    an_object.name = "update"
    inmemory_repository.update(an_object)

    assert len(inmemory_repository.entities) == 1
    assert list(inmemory_repository.entities.values())[0].name == "update"


def test_update_upsert(inmemory_repository, an_object):
    inmemory_repository.update(an_object, upsert=True)

    assert len(inmemory_repository.entities) == 1
    assert list(inmemory_repository.entities.values())[0] == an_object


def test_update_upsert_ignore(inmemory_repository, an_object):
    inmemory_repository.update(an_object)

    assert len(inmemory_repository.entities) == 0


def test_remove_one(inmemory_repository, an_object):
    inmemory_repository.add(an_object)

    from fractal.core.specifications.id_specification import IdSpecification

    inmemory_repository.remove_one(IdSpecification(an_object.id))

    assert len(inmemory_repository.entities) == 0


def test_find_one(inmemory_repository, an_object):
    inmemory_repository.add(an_object)

    from fractal.core.specifications.id_specification import IdSpecification

    assert inmemory_repository.find_one(IdSpecification(an_object.id)) == an_object


def test_find(inmemory_repository, an_object):
    inmemory_repository.add(an_object)

    assert len(list(inmemory_repository.find())) == 1


def test_find_with_specification(inmemory_repository, an_object):
    inmemory_repository.add(an_object)

    from fractal.core.specifications.id_specification import IdSpecification

    assert len(list(inmemory_repository.find(IdSpecification(an_object.id)))) == 1


def test_find_with_specification_empty(inmemory_repository, an_object):
    inmemory_repository.add(an_object)

    from fractal.core.specifications.id_specification import IdSpecification

    assert len(list(inmemory_repository.find(IdSpecification(2)))) == 0



from collections import defaultdict
from typing import List, Type

from fractal.core.command_bus.command import Command
from fractal.core.command_bus.command_handler import CommandHandler


class CommandBus(CommandHandler):
    def __init__(self):
        self.handlers = defaultdict(list)

    def commands(self) -> List[Type[Command]]:
        return []

    def set_handlers(self, handlers: List[CommandHandler]):
        for handler in handlers:
            for command in handler.commands():
                self.handlers[command.__name__].append(handler)

    def add_handler(self, handler: CommandHandler):
        for command in handler.commands():
            self.handlers[command.__name__].append(handler)

    async def handle_async(self, command: Command):
        ret = {}
        for handler in self.handlers[command.__class__.__name__]:
            res = await handler.handle(command)
            if res is not None:
                ret[handler.__class__.__name__] = res
        return ret

    def handle(self, command: Command):
        ret = {}
        for handler in self.handlers[command.__class__.__name__]:
            res = handler.handle(command)
            if res is not None:
                ret[handler.__class__.__name__] = res
        return ret



import logging
import os
from typing import List, Tuple


class ApplicationContext(object):
    instance = None

    def __new__(cls, dotenv=True, *args, **kwargs):
        if not isinstance(cls.instance, cls):
            cls.instance = object.__new__(cls, *args, **kwargs)
            if dotenv:
                from dotenv import load_dotenv

                load_dotenv()
            cls.instance.load()
        return cls.instance

    def load(self):
        from fractal.core.utils.loggers import init_logging

        init_logging(os.getenv("LOG_LEVEL", "INFO"))
        self.logger = logging.getLogger("app")
        self.repositories = []
        self.service_names = []

        self.load_internal_services()
        self.load_repositories()
        self.load_egress_services()

        from fractal.core.event_sourcing.event_publisher import EventPublisher

        self.event_publisher = EventPublisher(self.load_event_projectors())
        self.load_command_bus()
        self.load_ingress_services()

        for repository in self.repositories:
            assert repository.is_healthy()
        for service in self.services:
            assert service.is_healthy()

    def reload(self):
        self.load()

    def adapters(self):
        for repository in self.repositories:
            yield repository
        for service in self.services:
            yield service

    def load_internal_services(self):
        """Load services for internal use of the domain."""

    def load_repositories(self):
        """Load repositories for data access"""

    def load_egress_services(self):
        """Load services to external interfaces that are initiated by this service (outbound)"""

    def load_event_projectors(self):
        return []

    def load_command_bus(self):
        from fractal.core.command_bus.command_bus import CommandBus

        self.command_bus = CommandBus()

    def load_ingress_services(self):
        """Load services to external interfaces that are initiated by the external services (inbound)"""

    def install_repository(self, repository):
        self.repositories.append(repository)
        return repository

    def install_service(self, service, *, name=""):
        if not name:
            from fractal.core.utils.string import camel_to_snake

            name = camel_to_snake(service.__name__)
        self.service_names.append(name)
        setattr(
            ApplicationContext, name, property(lambda self: next(service.install(self)))
        )

    @property
    def services(self):
        services = []
        for service_name in self.service_names:
            services.append(getattr(self, service_name))
        return services

    def get_parameters(self, parameters: List[str]) -> Tuple:
        for parameter in parameters:
            if not hasattr(self, parameter):
                from fractal import FractalException

                raise FractalException(
                    f"ApplicationContext does not provide '{parameter}'"
                )
        return tuple(getattr(self, p) for p in parameters)



from abc import ABC, abstractmethod

from fractal.core.event_sourcing.event import BasicSendingEvent


class EventProjector(ABC):
    @abstractmethod
    def project(self, id: str, event: BasicSendingEvent):
        """Project the event, usually onto/into something defined in the constructor."""



from abc import ABC


class Service(ABC):
    @classmethod
    def install(cls, context):
        from fractal.core.utils.application_context import ApplicationContext

        assert issubclass(type(context), ApplicationContext)
        yield cls()

    def is_healthy(self) -> bool:
        return True



from fastapi import Depends
from fastapi.security import OAuth2PasswordBearer

from fractal import Fractal
from fractal.contrib.tokens.exceptions import NotAllowedException
from fractal.contrib.tokens.models import TokenPayload, TokenPayloadRoles

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="login")


def get_payload(fractal: Fractal, *, typ: str = "access"):
    assert hasattr(fractal, "context")
    context = getattr(fractal, "context")
    assert hasattr(context, "token_service")

    def _get_payload(token: str = Depends(oauth2_scheme)) -> TokenPayload:
        return TokenPayload(**context.token_service.verify(token, typ=typ))

    return _get_payload


def get_payload_roles(fractal: Fractal, *, roles: list, typ: str = "access"):
    assert hasattr(fractal, "context")
    context = getattr(fractal, "context")
    assert hasattr(context, "token_service")

    def _get_payload(token: str = Depends(oauth2_scheme)) -> TokenPayloadRoles:
        payload = TokenPayloadRoles(**context.token_service.verify(token, typ=typ))
        if not set(payload.roles or []) & set(roles):
            raise NotAllowedException("No permission!")
        return payload

    return _get_payload



from typing import Optional
from uuid import UUID

from pydantic.main import BaseModel


class TokenPayload(BaseModel):
    iss: Optional[str]
    sub: Optional[UUID]
    account: Optional[UUID]
    email: Optional[str]
    typ: Optional[str]


class TokenPayloadRoles(TokenPayload):
    roles: Optional[list]



def test_filter(inmemory_filter_repository, an_object, another_object):
    inmemory_filter_repository.add(an_object)
    inmemory_filter_repository.add(another_object)

    assert len(list(inmemory_filter_repository.find_filter("", fields=["name"]))) == 2
    assert len(list(inmemory_filter_repository.find_filter("default_name", fields=["name"]))) == 1
    assert len(list(inmemory_filter_repository.find_filter("default", fields=["name"]))) == 1
    assert len(list(inmemory_filter_repository.find_filter("name", fields=["name"]))) == 2
    assert len(list(inmemory_filter_repository.find_filter("t_n", fields=["name"]))) == 1
    assert len(list(inmemory_filter_repository.find_filter("_", fields=["name"]))) == 2
    assert len(list(inmemory_filter_repository.find_filter("x", fields=["name"]))) == 0


def test_filter_specification(inmemory_filter_repository, an_object, another_object):
    inmemory_filter_repository.add(an_object)
    inmemory_filter_repository.add(another_object)

    from fractal.core.specifications.id_specification import IdSpecification

    specification = IdSpecification("1")

    assert len(list(inmemory_filter_repository.find_filter("_", fields=["name"], specification=specification))) == 1


def test_filter_pre_processor(inmemory_filter_repository, an_object):
    an_object.name = an_object.name.upper()

    inmemory_filter_repository.add(an_object)

    assert len(list(inmemory_filter_repository.find_filter("default_name", fields=["name"]))) == 0
    assert len(list(inmemory_filter_repository.find_filter("DEFAULT_NAME", fields=["name"]))) == 1
    assert len(list(inmemory_filter_repository.find_filter("default_name", fields=["name"], pre_processor=lambda i: i.lower()))) == 1



from typing import Tuple

import pika
from pika import BlockingConnection, PlainCredentials
from pika.adapters.blocking_connection import BlockingChannel


def setup_rabbitmq_connection(
    host, port, username, password
) -> Tuple[BlockingConnection, BlockingChannel]:
    connection = pika.BlockingConnection(
        pika.ConnectionParameters(
            host=host,
            port=port,
            credentials=PlainCredentials(
                username=username,
                password=password,
            ),
        )
    )
    channel = connection.channel()
    return connection, channel



from typing import Optional

from fastapi import APIRouter, Depends

from fractal.contrib.fastapi.routers import Routes
from fractal.contrib.fastapi.routers.domain.models import AdapterInfo, Info
from fractal.contrib.fastapi.routers.tokens import get_payload
from fractal.contrib.tokens.fractal import DummyTokenServiceFractal
from fractal.contrib.tokens.models import TokenPayload
from fractal.core.utils.application_context import ApplicationContext
from fractal.core.utils.settings import Settings


def inject_default_routes(
    context: ApplicationContext, settings: Optional[Settings] = None
):
    router = APIRouter()

    @router.get(Routes.ROOT)
    def read_root():
        return {
            "FastAPI": settings.APP_NAME
            if settings and hasattr(settings, "APP_NAME")
            else "Fractal Service",
        }

    @router.get(Routes.INFO, responses={200: {"model": Info}})
    def info(
        payload: TokenPayload = Depends(
            get_payload(DummyTokenServiceFractal(context, settings))
        ),
    ):
        adapters = list(context.adapters())
        data = [
            AdapterInfo(
                adapter=adapter.__class__.__name__,
                status=False,
            )
            for adapter in adapters
        ]
        for i in range(len(data)):
            try:
                data[i].status_ok = adapters[i].is_healthy()
            except Exception as e:
                context.logger.error(e)
                data[i].status_ok = False
        return data

    return router



from dataclasses import make_dataclass

from fractal.core.specifications.generic.collections import (
    AndSpecification,
    OrSpecification,
)
from fractal.core.specifications.generic.operators import EqualsSpecification


def test_and_specification():
    spec = AndSpecification(
        [EqualsSpecification("id", 1), EqualsSpecification("name", "a")]
    )
    DC = make_dataclass("DC", [("id", int), ("name", str)])
    assert spec.is_satisfied_by(DC(**dict(id=1, name="a")))


def test_specification_and():
    spec = AndSpecification([EqualsSpecification("id", 1)]).And(
        EqualsSpecification("name", "a")
    )
    DC = make_dataclass("DC", [("id", int), ("name", str)])
    assert spec.is_satisfied_by(DC(**dict(id=1, name="a")))


def test_or_specification():
    spec = OrSpecification(
        [EqualsSpecification("id", 1), EqualsSpecification("name", "a")]
    )
    DC = make_dataclass("DC", [("id", int), ("name", str)])
    assert spec.is_satisfied_by(DC(**dict(id=1, name="b")))


def test_specification_or():
    spec = OrSpecification([EqualsSpecification("id", 1)]).Or(
        EqualsSpecification("name", "a")
    )
    DC = make_dataclass("DC", [("id", int), ("name", str)])
    assert spec.is_satisfied_by(DC(**dict(id=2, name="a")))



from typing import List, Type

import pytest


@pytest.fixture
def command_bus():
    from fractal.core.command_bus.command_bus import CommandBus

    return CommandBus()


@pytest.fixture
def command():
    from fractal.core.command_bus.command import Command

    return Command()


@pytest.fixture
def command_handler(command):
    from fractal.core.command_bus.command import Command
    from fractal.core.command_bus.command_handler import CommandHandler

    class FakeCommandHandler(CommandHandler):
        def commands(self) -> List[Type[Command]]:
            return [
                type(command),
            ]

        def handle(self, command: Command):
            return 1

    return FakeCommandHandler()


@pytest.fixture
def async_command_handler(command):
    from fractal.core.command_bus.command import Command
    from fractal.core.command_bus.command_handler import CommandHandler

    class FakeCommandHandler(CommandHandler):
        def commands(self) -> List[Type[Command]]:
            return [
                type(command),
            ]

        async def handle(self, command: Command):
            return 1

    return FakeCommandHandler()



import pytest


@pytest.fixture
def firestore_specification_builder():
    from fractal.contrib.gcp.firestore.specifications import (
        FirestoreSpecificationBuilder,
    )

    return FirestoreSpecificationBuilder



from tests.fixtures.core.command_bus import *  # NOQA NOSONAR
from tests.fixtures.core.event_sourcing import *  # NOQA NOSONAR
from tests.fixtures.core.fractal import *  # NOQA NOSONAR
from tests.fixtures.core.repositories import *  # NOQA NOSONAR
from tests.fixtures.core.services import *  # NOQA NOSONAR
from tests.fixtures.core.specifications import *  # NOQA NOSONAR
from tests.fixtures.core.utils import *  # NOQA NOSONAR



import json


def test_default_routes_root(fastapi_client):
    from fractal.contrib.fastapi.routers import Routes

    response = fastapi_client.get(Routes.ROOT)
    assert response.status_code == 200
    assert json.loads(response.content) == {"FastAPI": "Fractal Service"}


def test_default_routes_info(fastapi_client, token):
    from fractal.contrib.fastapi.routers import Routes

    response = fastapi_client.get(Routes.INFO, headers={"Authorization": f"Bearer {token}"})
    assert response.status_code == 200
    assert json.loads(response.content) == [
        {"adapter": "InMemoryRepository", "status_ok": True},
        {"adapter": "FakeService", "status_ok": True},
        {"adapter": "DummyJsonTokenService", "status_ok": True},
    ]


def test_default_routes_info_token_error(failing_service_fastapi_client):
    from fractal.contrib.fastapi.routers import Routes

    response = failing_service_fastapi_client.get(Routes.INFO)
    assert response.status_code == 401


def test_default_routes_info_error(failing_service_fastapi_client, token):
    from fractal.contrib.fastapi.routers import Routes

    response = failing_service_fastapi_client.get(Routes.INFO, headers={"Authorization": f"Bearer {token}"})
    assert response.status_code == 200
    assert json.loads(response.content) == [
        {"adapter": "InMemoryRepository", "status_ok": True},
        {"adapter": "FakeService", "status_ok": True},
        {"adapter": "DummyJsonTokenService", "status_ok": True},
        {"adapter": "FailingService", "status_ok": False},
    ]


def test_not_existing_route(failing_route_fastapi_client):
    from fractal.contrib.fastapi.routers import Routes

    response = failing_route_fastapi_client.get(Routes.INFO)
    assert response.status_code == 404


def test_failing_route(failing_route_fastapi_client):
    from fractal.contrib.fastapi.routers import Routes

    response = failing_route_fastapi_client.get(Routes.ROOT)
    assert response.status_code == 500
    assert json.loads(response.content) == {
        "code": "FAILING_ROUTE",
        "message": "failing_route",
    }



from dataclasses import make_dataclass

from fractal.core.specifications.generic.collections import AndSpecification
from fractal.core.specifications.generic.operators import (
    ContainsSpecification,
    EqualsSpecification,
    GreaterThanEqualSpecification,
    GreaterThanSpecification,
    InSpecification,
    LessThanEqualSpecification,
    LessThanSpecification,
)
from fractal.core.specifications.generic.specification import Specification


def test_parse():
    assert Specification.parse(id=1) == EqualsSpecification("id", 1)
    assert Specification.parse(id_x=1) == EqualsSpecification("id_x", 1)
    assert not Specification.parse(id__x=1)
    assert Specification.parse(id__equals=1) == EqualsSpecification("id", 1)
    assert Specification.parse(id__in=[1]) == InSpecification("id", [1])
    assert Specification.parse(id__contains="a") == ContainsSpecification("id", "a")
    assert Specification.parse(id__lt=1) == LessThanSpecification("id", 1)
    assert Specification.parse(id__lte=1) == LessThanEqualSpecification("id", 1)
    assert Specification.parse(id__gt=1) == GreaterThanSpecification("id", 1)
    assert Specification.parse(id__gte=1) == GreaterThanEqualSpecification("id", 1)
    assert Specification.parse(id=1, name="a") == AndSpecification(
        [EqualsSpecification("id", 1), EqualsSpecification("name", "a")]
    )
    assert Specification.parse(id__equals=1, name__equals="a") == AndSpecification(
        [EqualsSpecification("id", 1), EqualsSpecification("name", "a")]
    )
    assert Specification.parse(id__equals=1, name="a") == AndSpecification(
        [EqualsSpecification("id", 1), EqualsSpecification("name", "a")]
    )
    assert Specification.parse(id=1, name__equals="a") == AndSpecification(
        [EqualsSpecification("id", 1), EqualsSpecification("name", "a")]
    )
    assert Specification.parse(id__gt=1, name__contains="a") == AndSpecification(
        [GreaterThanSpecification("id", 1), ContainsSpecification("name", "a")]
    )


def test_specification_and():
    spec = EqualsSpecification("id", 1).And(EqualsSpecification("name", "a"))
    DC = make_dataclass("DC", [("id", int), ("name", str)])
    assert spec.is_satisfied_by(DC(**dict(id=1, name="a")))


def test_specification_or():
    spec = EqualsSpecification("id", 1).Or(EqualsSpecification("name", "a"))
    DC = make_dataclass("DC", [("id", int), ("name", str)])
    assert spec.is_satisfied_by(DC(**dict(id=1, name="a")))


def test_specification_not_and():
    spec = Specification.Not(
        EqualsSpecification("id", 1).And(EqualsSpecification("name", "a"))
    )
    DC = make_dataclass("DC", [("id", int), ("name", str)])
    assert spec.is_satisfied_by(DC(**dict(id=1, name="b")))


def test_specification_not_or():
    spec = Specification.Not(
        EqualsSpecification("id", 1).Or(EqualsSpecification("name", "a"))
    )
    DC = make_dataclass("DC", [("id", int), ("name", str)])
    assert spec.is_satisfied_by(DC(**dict(id=2, name="b")))



from typing import Dict, Iterator, Optional

from fractal.core.repositories import Entity, Repository
from fractal.core.specifications.generic.specification import Specification


class InMemoryRepositoryMixin(Repository[Entity]):
    def __init__(self):
        self.entities: Dict[str, Entity] = {}

    def add(self, entity: Entity) -> Entity:
        self.entities[entity.id] = entity
        return entity

    def update(self, entity: Entity, upsert=False) -> Entity:
        if entity.id in self.entities or upsert:
            return self.add(entity)

    def remove_one(self, specification: Specification):
        if obj := self.find_one(specification):
            del self.entities[obj.id]

    def find_one(self, specification: Specification) -> Optional[Entity]:
        for entity in filter(
            lambda i: specification.is_satisfied_by(i), self.entities.values()
        ):
            return entity

    def find(self, specification: Optional[Specification] = None) -> Iterator[Entity]:
        if specification:
            entities = filter(
                lambda i: specification.is_satisfied_by(i), self.entities.values()
            )
        else:
            entities = self.entities.values()
        for entity in entities:
            yield entity

    def is_healthy(self) -> bool:
        return True



import logging
from datetime import datetime

from google.cloud import pubsub_v1

from fractal.core.event_sourcing.event import BasicSendingEvent
from fractal.core.event_sourcing.event_projector import EventProjector
from fractal.core.event_sourcing.message import Message

logger = logging.getLogger("api")


class PubSubEventBusProjector(EventProjector):
    """
    https://cloud.google.com/pubsub/docs/quickstart-client-libraries#pubsub-client-libraries-python
    """

    def __init__(self, project_id):
        self.project_id = project_id
        self.publisher = pubsub_v1.PublisherClient()
        self.project_path = f"projects/{project_id}"

    def project(self, id: str, event: BasicSendingEvent):
        # The `topic_path` method creates a fully qualified identifier
        # in the form `projects/{project_id}/topics/{topic_id}`
        topic_path = self.publisher.topic_path(self.project_id, event.topic)

        # Check if topic exists
        if topic_path not in [
            t.name
            for t in self.publisher.list_topics(request={"project": self.project_path})
        ]:
            logger.error(f"Topic doesn't exist '{topic_path}'")

        # Wrap event in a message
        message = Message(
            id,
            datetime.utcnow(),
            event.__class__.__name__,
            event.to_json(),
        )

        # Data must be a bytestring
        data = message.to_json().encode("utf-8")

        # When you publish a message, the client returns a future.
        future = self.publisher.publish(topic_path, data=data)
        logger.debug(f"{data} sent to Google PubSub - message nr: {future.result()}")



import os

import pytest


@pytest.fixture
def settings_class():
    from fractal.core.utils.settings import Settings

    class FakeSettings(Settings):
        APP_NAME = os.getenv("APP_NAME", "app_name")

        def load(self):
            self.WEBSITE_HOST = os.getenv("WEBSITE_HOST", "http://localhost:8000")

    return FakeSettings


@pytest.fixture
def settings(settings_class):
    return settings_class()


@pytest.fixture
def fake_application_context_class(
    inmemory_repository, fake_service_class, another_fake_service_class
):
    from fractal.core.utils.application_context import ApplicationContext

    class FakeApplicationContext(ApplicationContext):
        def load_internal_services(self):
            if os.getenv("FAKE_SERVICE", "") == "another":
                self.install_service(another_fake_service_class, name="fake_service")
            else:
                self.install_service(fake_service_class)

        def load_repositories(self):
            self.install_repository(inmemory_repository)

        def load_egress_services(self):
            pass

        def load_event_projectors(self):
            return []

        def load_command_bus(self):
            super(FakeApplicationContext, self).load_command_bus()

        def load_ingress_services(self):
            pass

    return FakeApplicationContext


@pytest.fixture
def empty_application_context():
    from fractal.core.utils.application_context import ApplicationContext

    return ApplicationContext()



from typing import Any

from fractal.core.specifications.generic.operators import EqualsSpecification


class UserIdSpecification(EqualsSpecification):
    def __init__(self, user_id: Any):
        super(UserIdSpecification, self).__init__("user_id", user_id)



from abc import ABC, abstractmethod
from typing import Generic, Iterator, Optional, TypeVar

from fractal.core.specifications.generic.specification import Specification

Entity = TypeVar("Entity")


class Repository(Generic[Entity], ABC):
    @abstractmethod
    def add(self, entity: Entity) -> Entity:
        raise NotImplementedError

    @abstractmethod
    def update(self, entity: Entity, upsert=False) -> Entity:
        raise NotImplementedError

    @abstractmethod
    def remove_one(self, specification: Specification):
        raise NotImplementedError

    @abstractmethod
    def find_one(self, specification: Specification) -> Optional[Entity]:
        raise NotImplementedError

    @abstractmethod
    def find(self, specification: Specification = None) -> Iterator[Entity]:
        raise NotImplementedError

    @abstractmethod
    def is_healthy(self) -> bool:
        raise NotImplementedError



import pytest
from django.db.models import Q

specifications = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), {"id": 1}),
    (pytest.lazy_fixture("or_specification"), Q(id=1) | Q(name="test")),
    (pytest.lazy_fixture("and_specification"), Q(id=1, name="test")),
    (pytest.lazy_fixture("in_specification"), {"field__in": [1, 2, 3]}),
    (pytest.lazy_fixture("less_than_specification"), {"id__lt": 1}),
    (pytest.lazy_fixture("less_than_equal_specification"), {"id__lte": 1}),
    (pytest.lazy_fixture("greater_than_specification"), {"id__gt": 1}),
    (pytest.lazy_fixture("greater_than_equal_specification"), {"id__gte": 1}),
    (pytest.lazy_fixture("regex_string_match_specification"), {"id__regex": ".*abc.*"}),
    (pytest.lazy_fixture("dict_specification"), {"id": 1, "test": 2}),
]


@pytest.mark.parametrize("specification, expected", specifications)
def test_build(django_specification_builder, specification, expected):
    assert django_specification_builder.build(specification) == expected



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from fractal.core.command_bus.command import Command


@dataclass
class Event(object):
    pass


@dataclass
class BasicSendingEvent(Event, ABC):
    @property
    @abstractmethod
    def object_id(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def aggregate_root_id(self):
        raise NotImplementedError


@dataclass
class SendingEvent(BasicSendingEvent, ABC):
    command: Command

    @staticmethod
    def to_event(event: Any):
        return event


@dataclass
class ReceivingEvent(Event):
    def to_command(self):
        raise NotImplementedError



import logging

try:
    import sentry_sdk
    from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
except ImportError:
    sentry = False
else:
    sentry = True

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse

from fractal.contrib.fastapi.exceptions.error_message import ErrorMessage
from fractal.core.exceptions import DomainException
from fractal.core.utils.settings import Settings


def install_fastapi(settings: Settings):
    app = FastAPI(root_path=getattr(settings, "OPENAPI_PREFIX_PATH", ""))

    app.add_middleware(
        CORSMiddleware,
        allow_origins=getattr(settings, "ALLOW_ORIGINS", "").split(","),
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    if sentry:
        sentry_sdk.init(dsn=getattr(settings, "SENTRY_DSN", ""))
        app.add_middleware(
            SentryAsgiMiddleware,
        )

    @app.exception_handler(DomainException)
    def unicorn_domain_exception_handler(request: Request, exc: DomainException):
        logger = logging.getLogger("app")
        if isinstance(exc.status_code, int) and 400 <= exc.status_code < 500:
            logger.warning(f"{exc.code} - {exc.message}")
        else:
            logger.error(f"{exc.code} - {exc.message}")

        return JSONResponse(
            status_code=exc.status_code,
            content=ErrorMessage(
                code=exc.code,
                message=exc.message,
            ).dict(),
            headers=exc.headers,
        )

    @app.exception_handler(Exception)
    def unicorn_exception_handler(request: Request, exc: Exception):
        logging.error(exc)

        return JSONResponse(
            status_code=500,
            content=ErrorMessage(
                code=exc.__class__.__name__,
                message=str(exc),
            ).dict(),
        )

    return app



from dataclasses import asdict
from datetime import date
from decimal import Decimal
from typing import Generator, Optional

from google.cloud.firestore_v1 import Client

from fractal.contrib.gcp.firestore.specifications import FirestoreSpecificationBuilder
from fractal.core.repositories import Entity, Repository
from fractal.core.specifications.generic.specification import Specification


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class FirestoreRepositoryMixin(Repository[Entity]):
    """
    https://github.com/GoogleCloudPlatform/python-docs-samples/blob/46fa5a588858021ea32350584a4ee178cd7c1f33/firestore/cloud-client/snippets.py#L62-L66
    """

    entity = Entity

    def __init__(self, client: Client):
        self.collection = client.collection(self.entity.__name__.lower())

    def add(self, entity: Entity) -> Entity:
        doc_ref = self.collection.document(entity.id)
        doc_ref.set(asdict(entity))
        return entity

    def update(self, entity: Entity, *, upsert=False) -> Entity:
        if not upsert:
            doc_ref = self.collection.document(entity.id)
            doc = doc_ref.get()
            if doc.exists:
                doc_ref.set(asdict(entity))
            return entity
        return self.add(entity)

    def remove_one(self, specification: Specification):
        entity = self.find_one(specification)
        self.collection.document(entity.id).delete()

    def find_one(self, specification: Specification) -> Optional[Entity]:
        _filter = FirestoreSpecificationBuilder.build(specification)
        collection = self.collection
        if isinstance(_filter, list):
            for f in _filter:
                collection = collection.where(*f)
        else:
            collection = collection.where(*_filter)
        for doc in filter(
            lambda i: specification.is_satisfied_by(AttrDict(i.to_dict())),
            collection.stream(),
        ):
            return self.entity(**doc.to_dict())

    def find(
        self, specification: Specification = None
    ) -> Generator[Entity, None, None]:
        _filter = FirestoreSpecificationBuilder.build(specification)
        collection = self.collection
        if isinstance(_filter, list):
            for f in _filter:
                collection = collection.where(*f)
        else:
            collection = collection.where(*_filter)
        for doc in collection.stream():
            yield self.entity(**doc.to_dict())

    def is_healthy(self) -> bool:
        return True


class FirestoreDict(dict):
    def __init__(self, obj):
        for i, v in enumerate(obj):
            if type(v[1]) is Decimal:
                obj[i] = (v[0], f"{v[1]:.2f}")
            if type(v[1]) is date:
                obj[i] = (v[0], v[1].isoformat())
        super(FirestoreDict, self).__init__(obj)


class FirestoreRepositoryDictMixin(FirestoreRepositoryMixin[Entity]):
    def add(self, entity: Entity) -> Entity:
        doc_ref = self.collection.document(entity.id)
        doc_ref.set(asdict(entity, dict_factory=FirestoreDict))
        return entity

    def update(self, entity: Entity, *, upsert=False) -> Entity:
        if not upsert:
            doc_ref = self.collection.document(entity.id)
            doc = doc_ref.get()
            if doc.exists:
                doc_ref.set(asdict(entity, dict_factory=FirestoreDict))
            return entity
        return self.add(entity)



from typing import Iterable, Optional

from fractal.core.process.action import Action
from fractal.core.process.process import Process
from fractal.core.process.process_scope import ProcessScope
from fractal.core.specifications.generic.specification import Specification


class IfElseAction(Action):
    def __init__(
        self,
        specification: Specification,
        process_true: Process,
        process_false: Optional[Process] = None,
    ):
        self.specification = specification
        self.process_true = process_true
        self.process_false = process_false

    def execute(self, scope: ProcessScope) -> ProcessScope:
        if self.specification.is_satisfied_by(scope):
            scope.update(self.process_true.run(scope))
        elif self.process_false:
            scope.update(self.process_false.run(scope))
        return scope


class WhileAction(Action):
    def __init__(self, specification: Specification, process: Process):
        self.specification = specification
        self.process = process

    def execute(self, scope: ProcessScope) -> ProcessScope:
        while self.specification.is_satisfied_by(scope):
            scope.update(self.process.run(scope))
        return scope


class ForEachAction(Action):
    def __init__(self, iterable: Iterable, process: Process):
        self.iterable = iterable
        self.process = process

    def execute(self, scope: ProcessScope) -> ProcessScope:
        for item in self.iterable:
            scope["item"] = item
            scope.update(self.process.run(scope))
        return scope



"""Fractal is a scaffolding toolkit for building SOLID logic for your Python applications."""

__version__ = "0.2.9"

from abc import ABC


class FractalException(Exception):
    pass


class Fractal(ABC):
    def __init__(self):
        if not hasattr(self, "settings"):
            raise FractalException(
                "Fractal service doesn't have a 'settings' attribute."
            )
        if not hasattr(self, "context"):
            raise FractalException(
                "Fractal service doesn't have a 'context' attribute."
            )



import uuid
from dataclasses import dataclass

import pytest


@pytest.fixture
def account():
    @dataclass
    class Account:
        id: str

    return Account(str(uuid.uuid4()))


@pytest.fixture
def user(account):
    @dataclass
    class User:
        id: str
        account_id: str

    return User(str(uuid.uuid4()), account.id)


@pytest.fixture
def token(dummy_token_service_fractal, user, account):
    return dummy_token_service_fractal.context.token_service.generate(
        {
            "sub": user.id,
            "account": account.id,
        }
    )


@pytest.fixture
def admin_role_token(dummy_token_service_fractal, user, account):
    return dummy_token_service_fractal.context.token_service.generate(
        {
            "sub": user.id,
            "account": account.id,
            "roles": ["admin"],
        }
    )


@pytest.fixture
def token_service_application_context(fake_application_context_class, dummy_json_token_service_class):
    application_context = fake_application_context_class()
    application_context.install_service(dummy_json_token_service_class, name="token_service")
    return application_context



from typing import Any, Collection

from fractal.core.specifications.account_id_specification import AccountIdSpecification
from fractal.core.specifications.generic.collections import AndSpecification
from fractal.core.specifications.generic.specification import Specification
from fractal.core.specifications.id_specification import IdSpecification


class ObjectOfAccountSpecification(Specification):
    def __init__(self, object_id: Any, account_id: Any):
        self.specification = AndSpecification(
            [
                AccountIdSpecification(account_id),
                IdSpecification(object_id),
            ]
        )

    def is_satisfied_by(self, obj: Any) -> bool:
        return self.specification.is_satisfied_by(obj)

    def to_collection(self) -> Collection:
        return self.specification.to_collection()



import re


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def snake_to_camel(name):
    return "".join(word.title() for word in name.split("_"))



import pytest

data = [
    (pytest.lazy_fixture("a_base_model_object"), {"id": "1", "name": "default_name"}),
    ({1, 2, 3}, [1, 2, 3]),
]


@pytest.mark.parametrize("obj, expected", data)
def test_base_model_enhanced_encoder(obj, expected):
    from fractal.contrib.fastapi.utils.json_encoder import BaseModelEnhancedEncoder

    assert BaseModelEnhancedEncoder().default(obj) == expected



from tests.fixtures.contrib.tokens.fractal import *  # NOQA NOSONAR
from tests.fixtures.contrib.tokens.services import *  # NOQA NOSONAR
from tests.fixtures.contrib.tokens.utils import *  # NOQA NOSONAR



from copy import deepcopy
from typing import List

from fractal.core.event_sourcing.event import Event


class EventSourcedAggregateRoot:
    __events: List[Event] = []

    def record(self, event: Event):
        self.__events.append(event)
        return self

    def release(self):
        released_events = deepcopy(self.__events)
        self.__events.clear()
        return released_events



import pytest


def reset(dummy_token_service_fractal, settings):
    dummy_token_service_fractal.settings.reload(settings)
    dummy_token_service_fractal.context.reload()
    return dummy_token_service_fractal


@pytest.fixture
def dummy_token_service_fractal(token_service_application_context, settings):
    from fractal.contrib.tokens.fractal import DummyTokenServiceFractal

    return DummyTokenServiceFractal(token_service_application_context, settings)



from abc import ABC, abstractmethod
from typing import List, Type

from fractal.core.command_bus.command import Command


class CommandHandler(ABC):
    @abstractmethod
    def commands(self) -> List[Type[Command]]:
        """Returns list of commands that this handler can handle."""

    @abstractmethod
    def handle(self, command: Command):
        """Handle command, might return a value."""



from dataclasses import make_dataclass

from fractal.core.specifications.user_id_specification import UserIdSpecification


def test_user_id_specification():
    spec = UserIdSpecification("abc")
    DC = make_dataclass("DC", [("user_id", int)])
    assert spec.is_satisfied_by(DC(**dict(user_id="abc")))



import re
from functools import reduce
from typing import Collection, Optional, Union

from django.db.models import Q

from fractal.core.exceptions import DomainException
from fractal.core.specifications.generic.collections import (
    AndSpecification,
    OrSpecification,
)
from fractal.core.specifications.generic.operators import (
    EqualsSpecification,
    GreaterThanEqualSpecification,
    GreaterThanSpecification,
    InSpecification,
    LessThanEqualSpecification,
    LessThanSpecification,
    RegexStringMatchSpecification,
)
from fractal.core.specifications.generic.specification import Specification


class SpecificationNotMappedToDjangoOrm(DomainException):
    code = "SPECIFICATION_NOT_MAPPED_TO_DJANGO_ORM"
    status_code = 500


class DjangoOrmSpecificationBuilder:
    @staticmethod
    def create_q(filter):
        if type(filter) is list:
            return Q(*filter)
        elif type(filter) is dict:
            return Q(**filter)

    @staticmethod
    def build(specification: Specification = None) -> Optional[Union[Collection, Q]]:
        if specification is None:
            return None
        elif isinstance(specification, AndSpecification):
            return reduce(
                lambda x, y: x & y,
                [
                    DjangoOrmSpecificationBuilder.create_q(
                        DjangoOrmSpecificationBuilder.build(spec)
                    )
                    for spec in specification.to_collection()
                ],
            )
        elif isinstance(specification, OrSpecification):
            return reduce(
                lambda x, y: x | y,
                [
                    DjangoOrmSpecificationBuilder.create_q(
                        DjangoOrmSpecificationBuilder.build(spec)
                    )
                    for spec in specification.to_collection()
                ],
            )
        elif isinstance(specification, InSpecification):
            return {f"{specification.field}__in": specification.value}
        elif isinstance(specification, EqualsSpecification):
            return {specification.field: specification.value}
        elif isinstance(specification, LessThanSpecification):
            return {f"{specification.field}__lt": specification.value}
        elif isinstance(specification, LessThanEqualSpecification):
            return {f"{specification.field}__lte": specification.value}
        elif isinstance(specification, GreaterThanSpecification):
            return {f"{specification.field}__gt": specification.value}
        elif isinstance(specification, GreaterThanEqualSpecification):
            return {f"{specification.field}__gte": specification.value}
        elif isinstance(specification, RegexStringMatchSpecification):
            return {
                f"{specification.field}__regex": rf".*{re.escape(specification.value)}.*"
            }
        elif isinstance(specification.to_collection(), dict):
            return specification.to_collection()
        raise SpecificationNotMappedToDjangoOrm(
            f"Specification '{specification}' not mapped to Django Orm query."
        )



import pytest

specifications = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), ("id", "==", 1)),
    (
        pytest.lazy_fixture("and_specification"),
        [("id", "==", 1), ("name", "==", "test")],
    ),
    (pytest.lazy_fixture("dict_specification"), [("id", "==", 1), ("test", "==", 2)]),
]


@pytest.mark.parametrize("specification, expected", specifications)
def test_build(firestore_specification_builder, specification, expected):
    assert firestore_specification_builder.build(specification) == expected
